# Copyright Karl Otness
# SPDX-License-Identifier: MIT


import powerpax as ppx
import jax


def test_roundtrip_static():
    obj = ppx.Static(1.5)
    flat, tree = jax.tree_util.tree_flatten(obj)
    obj2 = jax.tree_util.tree_unflatten(tree, flat)
    assert obj == obj2
    assert hash(obj) == hash(obj2)


def test_no_leaves():
    obj = ppx.Static(1.5)
    leaves = jax.tree_util.tree_leaves(obj)
    assert len(leaves) == 0



# Copyright Karl Otness
# SPDX-License-Identifier: MIT


import operator
import functools
import typing
import jax
import jax.numpy as jnp


C = typing.TypeVar("C", bound=typing.Callable[..., typing.Any])
A = typing.TypeVar("A", bound=tuple[object, ...])
K = typing.TypeVar("K", bound=dict[str, object])


def get_leaf_length(leaf: typing.Any) -> int:
    if jnp.ndim(leaf) < 1:
        raise ValueError("attempted to vmap over array with zero dimensions")
    return operator.index(jnp.shape(leaf)[0])


def determine_splits(
    args: tuple[typing.Any, ...], kwargs: dict[str, typing.Any], chunk_size: int
) -> tuple[int, int]:
    # Determine leading axis size
    lead_sizes = {
        get_leaf_length(leaf)
        for leaf in jax.tree_util.tree_leaves((args, list(kwargs.values())))
    }
    if not lead_sizes:
        raise ValueError("vmap over empty arguments")
    elif len(lead_sizes) > 1:
        lead_size_str = ", ".join(map(str, lead_sizes))
        raise ValueError(f"inconsistent leading dimensions for vmap: {lead_size_str}")
    num_chunks, remainder = divmod(lead_sizes.pop(), operator.index(chunk_size))
    return num_chunks, remainder


def split_args(
    args: A, kwargs: K, chunk_size: int, num_chunks: int, remainder: int
) -> tuple[tuple[A, K], typing.Union[tuple[A, K], tuple[None, None]]]:
    main_size = num_chunks * chunk_size
    leading_args, leading_kwargs = jax.tree_util.tree_map(
        lambda arr: arr[:main_size].reshape((num_chunks, chunk_size) + arr.shape[1:]),
        (args, kwargs),
    )
    # Remainder
    if remainder < 1:
        rem_args = None
        rem_kwargs = None
    else:
        rem_args, rem_kwargs = jax.tree_util.tree_map(
            operator.itemgetter(slice(-remainder, None)), (args, kwargs)
        )
    return (leading_args, leading_kwargs), (rem_args, rem_kwargs)


def chunked_vmap(fun: C, chunk_size: int) -> C:
    r"""Like :func:`jax.vmap` but limited to batches of `chunk_size`
    steps.

    This function behaves like :func:`jax.vmap` in that it vectorizes
    a function to apply over batches of inputs. However, unlike vmap
    which carries out all calculations in parallel, this version will
    perform at most `chunk_size` steps at once and uses
    :func:`jax.lax.scan` to loop over chunks.

    This is useful in cases where the calculations in `fun` involve
    large intermediate values which can exhaust available memory. With
    `chunked_vmap` it is possible to place an upper bound on peak
    memory use from the intermediate results while preserving some of
    the performance benefits of vmap, particularly on GPUs.

    Parameters
    ----------
    fun: function
        Function to be mapped over additional axes.

    chunk_size: int
        Upper limit on the size of chunks to be vectorized over.
        Inputs larger than this will be processed with an outer
        :func:`scan <jax.lax.scan>` loop.

    Returns
    -------
    function
        A wrapped version of `fun` which vectorizes over leading axes
        of each input.

    Note
    ----
    Unlike :func:`jax.vmap` this function does not allow specifying
    `in_axes` or `out_axes`. It vectorizes all parameters over the
    first axis. Use :func:`jnp.moveaxis <jax.numpy.moveaxis>` and
    :func:`functools.partial` to map over other axes or to leave a
    parameter un-vectorized.
    """
    chunk_size = operator.index(chunk_size)
    if chunk_size <= 0:
        raise ValueError(f"chunk_size must be positive (got {chunk_size})")

    def outer_scan(_, x):
        args, kwargs = x
        ret = jax.vmap(fun)(*args, **kwargs)
        return None, ret

    @functools.wraps(fun)
    def wrapped_fun(*args, **kwargs):
        num_chunks, remainder = determine_splits(args, kwargs, chunk_size)
        if num_chunks < 1 or (num_chunks == 1 and remainder == 0):
            # Trivial case: only one vmap needed
            return jax.vmap(fun)(*args, **kwargs)
        # Divide arguments into part for a scan, and a remainder chunk
        (leading_args, leading_kwargs), (rem_args, rem_kwargs) = split_args(
            args, kwargs, chunk_size, num_chunks, remainder
        )
        # Perform a scan for the main chunks
        _, ret = jax.lax.scan(
            outer_scan, None, (leading_args, leading_kwargs), length=num_chunks
        )
        ret = jax.tree_util.tree_map(
            lambda arr: arr.reshape((-1,) + arr.shape[2:]), ret
        )
        # Next, the remainder (if needed)
        if rem_args is not None and rem_kwargs is not None:
            ret = jax.tree_util.tree_map(
                lambda a, b: jnp.concatenate([a, b], axis=0),
                ret,
                jax.vmap(fun)(*rem_args, **rem_kwargs),
            )
        return ret

    return typing.cast(C, wrapped_fun)



# Copyright Karl Otness
# SPDX-License-Identifier: MIT


import operator
import functools
import typing
import jax
import jax.numpy as jnp


C = typing.TypeVar("C")
X = typing.TypeVar("X")
Y = typing.TypeVar("Y")


def get_leaf_length(leaf: jax.Array) -> int:
    if jnp.ndim(leaf) < 1:
        raise ValueError("pytree array leaf is zero-dimensional, cannot scan")
    return operator.index(jnp.shape(leaf)[0])


def get_target_length(xs: object, length: typing.Optional[int]) -> int:
    leaf_lengths = {get_leaf_length(x) for x in jax.tree_util.tree_leaves(xs)}
    if len(leaf_lengths) > 1:
        size_str = ", ".join(map(str, leaf_lengths))
        raise ValueError(f"inconsistent lengths for tree input: {size_str}")
    if length is not None:
        length = operator.index(length)
        if length < 0:
            raise ValueError(f"invalid length {length}, must not be negative")
        leaf_lengths.add(length)
    if not leaf_lengths:
        raise ValueError("no input values provided and no specified length")
    elif len(leaf_lengths) > 1:
        size_str = ", ".join(map(str, leaf_lengths))
        raise ValueError(f"ambiguous lengths for scan: {size_str}")
    return leaf_lengths.pop()


def clip(a: int, a_min: int, a_max: int) -> int:
    return min(max(a, a_min), a_max)


def compute_slices(
    start: int, step: int, num_ys: int, target_length: int, reverse: bool
) -> tuple[slice, typing.Optional[int], slice, int, slice]:
    # First, compute the start and end indices of the iteration range (inclusive)
    start_index = start
    end_index = start + (step * (num_ys - 1))
    # Next, we compute how many steps we skip at the left/right side of the array
    if step > 0:
        # Looping in increasing order
        start_skip = start_index
        end_skip = target_length - end_index - 1
    else:
        # Looping in decreasing order (we have to skip end_index steps steps)
        start_skip = end_index
        end_skip = target_length - start_index - 1
    # If the loop iterates in reverse, we have to skip the right side steps first
    if reverse:
        start_skip, end_skip = end_skip, start_skip
    # Determine how many iterations we have to do for the core loop (outputs ys)
    if num_ys == 1:
        # Special case for 1 output (do the core without a scan)
        pre_step = True
        outer_steps = 0
    elif start_skip < abs(step) - 1:
        # We have to separate out the first step
        pre_step = True
        outer_steps = num_ys - 1
    else:
        # We can merge the first step with part of the start skip
        start_skip -= abs(step) - 1
        pre_step = False
        outer_steps = num_ys
    # Compute slices to divide xs into chunks
    if reverse:
        # Looping in reverse order (start_slice is at the right end of the array)
        start_slice = slice(target_length - start_skip, None)
        pre_slice = target_length - start_skip - 1 if pre_step else None
        core_slice = slice(
            end_skip, target_length - start_skip - (1 if pre_step else 0)
        )
        end_slice = slice(None, end_skip)
    else:
        # Looping in forward order (start_slice is at the left end of the array)
        start_slice = slice(None, start_skip)
        pre_slice = start_skip if pre_step else None
        core_slice = slice(
            start_skip + (1 if pre_step else 0), target_length - end_skip
        )
        end_slice = slice(target_length - end_skip, None)
    return start_slice, pre_slice, core_slice, outer_steps, end_slice


def sliced_scan(
    f: typing.Callable[[C, X], tuple[C, Y]],
    init: C,
    xs: X,
    length: typing.Optional[int] = None,
    reverse: bool = False,
    unroll: int = 1,
    *,
    start: typing.Optional[int] = None,
    stop: typing.Optional[int] = None,
    step: typing.Optional[int] = None,
) -> tuple[C, Y]:
    r"""Slice the output of :func:`jax.lax.scan` without first
    collecting all iterations.

    Using this function is equivalent to::

       carry, ys = jax.lax.scan(f, init, xs, length, reverse, unroll)
       ys = jax.tree_util.tree_map(
           lambda leaf: leaf[start:stop:step], ys
       )

    except that it *does not* first produce a complete `ys`.
    Internally the loop is split into several separate (and in some
    cases nested) scan phases to collect only the required steps. See
    :func:`jax.tree_util.tree_map` for information on its effect in
    the above example.

    Most arguments are as in :func:`jax.lax.scan`. New parameters for
    this function are `start`, `stop`, and `step`.

    Parameters
    ----------
    f: function
        A function suitable for use with :func:`jax.lax.scan`. Namely,
        takes two arguments (a `carry` and `x`) and returns two values
        (an updated `carry` and `y`).

    init: object
        A JAX pytree initializing the carry.

    xs: object
        A JAX pytree over which to loop. If not :pycode:`None` the
        loop scans over the leading dimension of each leaf
        :class:`array <jax.Array>`.

    length: int, optional
        Integer specifying the number of iterations. Useful if `xs` is
        :pycode:`None`. If both `length` and `xs` are provided, the
        implied loop iteration counts must match.

    reverse: bool, optional
        If :pycode:`False` (default) the loop proceeds in normal
        forward order. Otherwise the loop will start at the end of
        each input array in `xs` and fill `ys` from right to left.

    unroll: int, optional
        An integer allowing greater loop unrolling. Note that this
        function applies the unrolling to each internal :func:`scan
        <jax.lax.scan>` adjusting so that this provides an upper bound
        on the number of unrolled steps for the innermost loop in the
        case of nested scans.

    start: int, optional
        The starting index at which the slice should start.

    stop: int, optional
        The ending index at which the slice should stop.

    step: int, optional
        The step size for the slice.

    Returns
    -------
    object, object
        A tuple :pycode:`(carry, ys)` where `ys` has been sliced by
        `start`, `stop` and `step`.

    Note
    ----
    The slicing applies only to `ys`, the carry value is still updated
    by all loop iterations even if their `y` outputs are skipped by
    the slice.
    """

    def skip_step(carry, x):
        new_carry, _ = f(carry, x)
        return new_carry, None

    target_length = get_target_length(xs, length)
    unroll = clip(operator.index(unroll), 1, target_length)
    start, stop, step = slice(start, stop, step).indices(target_length)
    num_ys = len(range(start, stop, step))

    if num_ys == target_length:
        # Normal scan
        carry, ys = jax.lax.scan(
            f,
            init,
            xs,
            length=target_length,
            reverse=reverse,
            unroll=unroll,
        )
        if step < 0:
            ys = jax.tree_util.tree_map(functools.partial(jnp.flip, axis=0), ys)
        return carry, ys
    elif num_ys == 0:
        # No core iterations. Loop for the carry and produce empty ys
        carry, _ = jax.lax.scan(
            skip_step,
            init,
            xs,
            length=target_length,
            reverse=reverse,
            unroll=unroll,
        )

        def dummy_scan(init, xs):
            _, ys = jax.lax.scan(
                f, init, xs, length=length, reverse=reverse, unroll=unroll
            )
            return ys

        ys = jax.tree_util.tree_map(
            lambda sd: jnp.zeros((0,) + sd.shape[1:], dtype=sd.dtype),
            jax.eval_shape(dummy_scan, init, xs),
        )
        return carry, ys

    def scan_y_to_carry(carry, x):
        old_carry, _ = carry
        new_carry, new_y = f(old_carry, x)
        return (new_carry, new_y), None

    def inner_scan(carry, xs):
        dummy_y = jax.tree_util.tree_map(
            lambda sd: jnp.zeros(sd.shape, dtype=sd.dtype),
            jax.eval_shape(
                lambda carry, xs: f(
                    carry, jax.tree_util.tree_map(operator.itemgetter(0), xs)
                )[1],
                carry,
                xs,
            ),
        )
        (carry, y), _ = jax.lax.scan(
            scan_y_to_carry,
            (carry, dummy_y),
            xs,
            length=abs(step),
            reverse=reverse,
            unroll=clip(unroll, 1, abs(step)),
        )
        return carry, y

    start_slice, pre_slice, core_slice, outer_steps, end_slice = compute_slices(
        start=start,
        step=step,
        num_ys=num_ys,
        target_length=target_length,
        reverse=reverse,
    )

    def do_skip_steps(carry: C, slicer: slice) -> C:
        length = len(range(*slicer.indices(target_length)))
        if length <= 0:
            new_carry = carry
        elif length == 1:
            # Call function directly
            new_carry, _ = jax.jit(skip_step)(
                carry, jax.tree_util.tree_map(operator.itemgetter(slicer.start), xs)
            )
        else:
            new_carry, _ = jax.lax.scan(
                skip_step,
                carry,
                jax.tree_util.tree_map(operator.itemgetter(slicer), xs),
                length=length,
                reverse=reverse,
                unroll=clip(unroll, 1, length),
            )
        return new_carry

    carry = do_skip_steps(init, start_slice)
    if pre_slice is not None:
        carry, pre_y = jax.jit(f)(
            carry, jax.tree_util.tree_map(operator.itemgetter(pre_slice), xs)
        )
        pre_y = jax.tree_util.tree_map(
            functools.partial(jnp.expand_dims, axis=0), pre_y
        )
    else:
        pre_y = None
    # Do core steps
    if outer_steps > 0:
        if abs(step) > 1:
            # Requires a nested scan
            def leaf_reshape(leaf):
                return leaf[core_slice].reshape(
                    (outer_steps, abs(step)) + leaf.shape[1:]
                )

            carry, ys = jax.lax.scan(
                inner_scan,
                carry,
                jax.tree_util.tree_map(leaf_reshape, xs),
                length=outer_steps,
                reverse=reverse,
                unroll=clip(unroll // abs(step), 1, outer_steps),
            )
        else:
            # No nested scan required
            carry, ys = jax.lax.scan(
                f,
                carry,
                jax.tree_util.tree_map(operator.itemgetter(core_slice), xs),
                length=outer_steps,
                reverse=reverse,
                unroll=clip(unroll, 1, outer_steps),
            )
        # Concatenate if necessary
        if pre_y is not None:
            ys = jax.tree_util.tree_map(
                lambda a, b: jnp.concatenate([a, b] if not reverse else [b, a], axis=0),
                pre_y,
                ys,
            )
    else:
        ys = typing.cast(Y, pre_y)
    # Do end steps
    carry = do_skip_steps(carry, end_slice)
    if step < 0:
        ys = jax.tree_util.tree_map(functools.partial(jnp.flip, axis=0), ys)
    return carry, ys


def checkpoint_chunked_scan(
    f: typing.Callable[[C, X], tuple[C, Y]],
    init: C,
    xs: X,
    length: typing.Optional[int] = None,
    reverse: bool = False,
    unroll: int = 1,
    *,
    chunk_size: typing.Optional[int] = None,
) -> tuple[C, Y]:
    r"""Perform a :func:`scan <jax.lax.scan>` inserting
    :func:`checkpoints <jax.checkpoint>` every `chunk_size` steps.

    This function performs a normal scan loop, but inserts checkpoints
    at regular intervals. This can reduce peak memory use (at the cost
    of recomputation) when computing gradients through the loop.

    Most arguments are as in :func:`jax.lax.scan`. This function has
    one added parameter, `chunk_size`.

    Parameters
    ----------
    f: function
        A function suitable for use with :func:`jax.lax.scan`. Namely,
        takes two arguments (a carry and `x`) and returns two values
        (an updated carry and `y`).

    init: object
        A JAX pytree initializing the carry.

    xs: object
        A JAX pytree over which to loop. If not :pycode:`None` the
        loop scans over the leading dimension of each leaf
        :class:`array <jax.Array>`.

    length: int, optional
        Integer specifying the number of iterations. Useful if `xs` is
        :pycode:`None`. If both `length` and `xs` are provided, the
        implied loop iteration counts must match.

    reverse: bool, optional
        If :pycode:`False` (default) the loop proceeds in normal
        forward order. Otherwise the loop will start at the end of
        each input array in `xs` and fill `ys` from right to left.

    unroll: int, optional
        An integer allowing greater loop unrolling. Note that this
        function applies the unrolling to each internal :func:`scan
        <jax.lax.scan>` adjusting so that this provides an upper bound
        on the number of unrolled steps for the innermost loop in the
        case of nested scans.

    chunk_size: int, optional
        The interval at which to insert checkpoints. Every
        `chunk_size` steps a checkpoint will be inserted, starting
        with the first step. If this parameter is not specified, the
        entire scan is treated as one chunk with one checkpoint at the
        start.

    Returns
    -------
    object, object
        A tuple :pycode:`(carry, ys)`.
    """
    jax_checkpoint = jax.checkpoint  # type: ignore[attr-defined]
    target_length = get_target_length(xs, length)
    unroll = clip(operator.index(unroll), 1, target_length)
    if chunk_size is None:
        chunk_size = max(1, target_length)
    chunk_size = operator.index(chunk_size)
    if chunk_size <= 0:
        raise ValueError(f"chunk_size must be positive, but got {chunk_size}")
    if target_length <= chunk_size:
        # Do a normal scan with one checkpoint around it
        return typing.cast(
            tuple[C, Y],
            jax_checkpoint(
                lambda init, xs: jax.lax.scan(
                    f, init, xs, length=length, reverse=reverse, unroll=unroll
                )
            )(init, xs),
        )
    num_chunks, rem_steps = divmod(target_length, chunk_size)
    if rem_steps > 0:
        remainder = jax.tree_util.tree_map(
            operator.itemgetter(
                slice(-rem_steps, None) if not reverse else slice(None, rem_steps)
            ),
            xs,
        )
    else:
        remainder = None
    core_steps = num_chunks * chunk_size
    core = jax.tree_util.tree_map(
        lambda leaf: (leaf[:core_steps] if not reverse else leaf[-core_steps:]).reshape(
            (num_chunks, chunk_size)
        ),
        xs,
    )
    carry = init

    @functools.partial(jax_checkpoint, prevent_cse=False)
    def inner_scan(carry, xs):
        return jax.lax.scan(
            f,
            carry,
            xs,
            length=chunk_size,
            reverse=reverse,
            unroll=clip(unroll, 1, chunk_size),
        )

    carry, ys = jax.lax.scan(
        inner_scan,
        carry,
        core,
        length=num_chunks,
        reverse=reverse,
        unroll=clip(unroll // chunk_size, 1, num_chunks),
    )
    ys = jax.tree_util.tree_map(
        lambda leaf: leaf.reshape((core_steps,) + leaf.shape[2:]), ys
    )
    if rem_steps > 0:
        carry, rem_ys = jax_checkpoint(
            lambda init, xs: jax.lax.scan(
                f,
                init,
                xs,
                length=rem_steps,
                reverse=reverse,
                unroll=clip(unroll, 1, rem_steps),
            )
        )(carry, remainder)
        ys = jax.tree_util.tree_map(
            lambda a, b: jnp.concatenate([a, b] if not reverse else [b, a], axis=0),
            ys,
            rem_ys,
        )

    return carry, ys



# Copyright Karl Otness
# SPDX-License-Identifier: MIT


import dataclasses
import typing
import jax


T = typing.TypeVar("T")


@dataclasses.dataclass(frozen=True)
class Static(typing.Generic[T]):
    r"""Treat a value as :term:`jax:static` when processing a pytree.

    This class wraps `value` and will instruct JAX to treat it as a
    static value during pytree processing.

    Parameters
    ----------
    value: object
        The object to wrap and treat as static (must be
        :term:`python:hashable`).

    Attributes
    ----------
    value: object
        The wrapped, static value.
    """

    value: T


jax.tree_util.register_pytree_node(
    Static, lambda node: ((), node.value), lambda aux, _: Static(aux)
)



# Copyright Karl Otness
# SPDX-License-Identifier: MIT


import operator
import jax
import jax.numpy as jnp
import powerpax as ppx
import pytest


@pytest.mark.parametrize(
    "start,stop,step",
    [
        (None, None, None),
        (None, None, -1),
        (5, None, None),
        (None, -3, None),
        (-1, 0, 1),
        (0, -1, -1),
        (None, None, 2),
        (1, None, 2),
        (2, None, 2),
        (1, -3, 3),
        (-7, -2, 2),
        (2, 3, 4),
        (3, 2, -4),
    ],
)
@pytest.mark.parametrize("reverse", [False, True])
@pytest.mark.parametrize(
    "use_xs,use_length", [(True, True), (True, False), (False, True)]
)
def test_matches_scan(start, stop, step, reverse, use_xs, use_length):
    def scan_fn(carry, x):
        return carry + 1, (carry, x)

    length = 15
    xs = jnp.arange(length) if use_xs else None
    extra_args = {"length": length} if use_length else {}
    jax_carry, jax_ys = jax.lax.scan(scan_fn, 0, xs, reverse=reverse, **extra_args)
    jax_ys = jax.tree_util.tree_map(
        operator.itemgetter(slice(start, stop, step)), jax_ys
    )
    ppx_carry, ppx_ys = jax.jit(
        lambda init, xs: ppx.sliced_scan(
            scan_fn,
            init,
            xs,
            reverse=reverse,
            start=start,
            stop=stop,
            step=step,
            **extra_args,
        )
    )(0, xs)
    assert jax.tree_util.tree_all(
        jax.tree_util.tree_map(lambda la, lb: jnp.all(la == lb), jax_ys, ppx_ys)
    )
    assert jax.tree_util.tree_all(
        jax.tree_util.tree_map(lambda la, lb: jnp.all(la == lb), jax_carry, ppx_carry)
    )



# Copyright Karl Otness
# SPDX-License-Identifier: MIT


import jax
import jax.numpy as jnp
import powerpax as ppx
import pytest


@pytest.mark.parametrize("chunk_size", [1, 3, 5, 9, 10, 11])
@pytest.mark.parametrize(
    "use_args,use_kwargs", [(True, True), (True, False), (False, True)]
)
def test_matches_vmap(chunk_size, use_args, use_kwargs):
    def fun(arg=1, arg2=1):
        return arg * arg2

    arr = jnp.arange(10)
    args = (arr,) if use_args else ()
    kwargs = {"arg2": arr} if use_kwargs else {}
    jax_res = jax.vmap(fun)(*args, **kwargs)
    chunked_res = jax.jit(ppx.chunked_vmap(fun, chunk_size=chunk_size))(*args, **kwargs)
    assert jax_res.dtype == chunked_res.dtype
    assert jax_res.shape == chunked_res.shape
    assert jnp.all(jax_res == chunked_res)


@pytest.mark.parametrize("chunk_size", [-1, 0])
def test_error_invalid_chunk_size(chunk_size):
    def fun(arg):
        return arg * 2

    with pytest.raises(ValueError):
        _ = ppx.chunked_vmap(fun, chunk_size=chunk_size)



# Copyright Karl Otness
# SPDX-License-Identifier: MIT


import jax
import jax.numpy as jnp
import powerpax as ppx
import pytest


@pytest.mark.parametrize("chunk_size", [None, 1, 2, 5, 7, 15, 100])
@pytest.mark.parametrize("reverse", [True, False])
@pytest.mark.parametrize(
    "use_xs,use_length", [(True, True), (True, False), (False, True)]
)
def test_matches_scan(chunk_size, reverse, use_xs, use_length):
    def scan_fn(carry, x):
        return carry + 1, (carry, x)

    length = 15
    xs = jnp.arange(length) if use_xs else None
    extra_args = {"length": length} if use_length else {}
    jax_carry, jax_ys = jax.lax.scan(scan_fn, 0, xs, reverse=reverse, **extra_args)
    ppx_carry, ppx_ys = jax.jit(
        lambda init, xs: ppx.checkpoint_chunked_scan(
            scan_fn, init, xs, reverse=reverse, chunk_size=chunk_size, **extra_args
        )
    )(0, xs)
    assert jax.tree_util.tree_all(
        jax.tree_util.tree_map(lambda la, lb: jnp.all(la == lb), jax_ys, ppx_ys)
    )
    assert jax.tree_util.tree_all(
        jax.tree_util.tree_map(lambda la, lb: jnp.all(la == lb), jax_carry, ppx_carry)
    )



# Copyright Karl Otness
# SPDX-License-Identifier: MIT

r"""Useful utilities for JAX.

This package provides a small set of useful utilities for working with
JAX. For example, functions to limit the memory consumption of
:func:`jax.vmap` (:func:`ppx.chunked_vmap <powerpax.chunked_vmap>`) or
to keep the outputs of only a subset of :func:`jax.lax.scan`
iterations (:func:`ppx.sliced_scan <powerpax.sliced_scan>`).
"""

__version__ = "0.1.2"
__all__ = ["chunked_vmap", "sliced_scan", "checkpoint_chunked_scan", "Static"]

from ._vmap import chunked_vmap
from ._loop import sliced_scan, checkpoint_chunked_scan
from ._tree import Static
